@@ -25,7 +25,7 @@ void arraylist_free(arraylist_t *a) JL_NOTSAFEPOINT;
 
 void arraylist_push(arraylist_t *a, void *elt) JL_NOTSAFEPOINT;
 void *arraylist_pop(arraylist_t *a) JL_NOTSAFEPOINT;
-void arraylist_grow(arraylist_t *a, size_t n) JL_NOTSAFEPOINT;
+JL_DLLEXPORT void arraylist_grow(arraylist_t *a, size_t n) JL_NOTSAFEPOINT;
 
 typedef struct {
     uint32_t len;
@@ -39,7 +39,7 @@ void small_arraylist_free(small_arraylist_t *a) JL_NOTSAFEPOINT;
 
 void small_arraylist_push(small_arraylist_t *a, void *elt) JL_NOTSAFEPOINT;
 void *small_arraylist_pop(small_arraylist_t *a) JL_NOTSAFEPOINT;
-void small_arraylist_grow(small_arraylist_t *a, uint32_t n) JL_NOTSAFEPOINT;
+JL_DLLEXPORT void small_arraylist_grow(small_arraylist_t *a, uint32_t n) JL_NOTSAFEPOINT;
 
 #ifdef __cplusplus
 }
