@@ -91,9 +91,6 @@ end
         HelloWorldC_exe_path = joinpath(HelloWorldC_dir, "bin", "hello_world$(exeext)")
         @test isfile(HelloWorldC_exe_path)
 
-        HelloWorldC_dir_explicit_artifact = eval(:(@artifact_str "HelloWorldC" nothing joinpath(@__DIR__, "Artifacts.toml")))
-        @test isdir(HelloWorldC_dir_explicit_artifact)
-
         # Simple slash-indexed lookup
         HelloWorldC_bin_path = artifact"HelloWorldC/bin"
         @test isdir(HelloWorldC_bin_path)
@@ -164,6 +161,6 @@ end
 @testset "`Artifacts.artifact_names` and friends" begin
     n = length(Artifacts.artifact_names)
     @test length(Base.project_names) == n
-    @test length(Base.manifest_names) == n
+    @test length(Base.manifest_names) == 2n # there are two manifest names per project name
     @test length(Base.preferences_names) == n
 end
