@@ -16,20 +16,25 @@ export
     FileMonitor,
     FolderMonitor,
     PollingFileWatcher,
-    FDWatcher
+    FDWatcher,
+    # pidfile:
+    mkpidlock,
+    trymkpidlock
 
 import Base: @handle_as, wait, close, eventloop, notify_error, IOError,
     _sizeof_uv_poll, _sizeof_uv_fs_poll, _sizeof_uv_fs_event, _uv_hook_close, uv_error, _UVError,
     iolock_begin, iolock_end, associate_julia_struct, disassociate_julia_struct,
-    preserve_handle, unpreserve_handle, isreadable, iswritable, |
+    preserve_handle, unpreserve_handle, isreadable, iswritable, isopen,
+    |, getproperty, propertynames
 import Base.Filesystem.StatStruct
 if Sys.iswindows()
     import Base.WindowsRawSocket
 end
 
+
 # libuv file watching event flags
-const UV_RENAME = 1
-const UV_CHANGE = 2
+const UV_RENAME = Int32(1)
+const UV_CHANGE = Int32(2)
 struct FileEvent
     renamed::Bool
     changed::Bool
@@ -45,33 +50,36 @@ FileEvent(flags::Integer) = FileEvent((flags & UV_RENAME) != 0,
               a.changed | b.changed,
               a.timedout | b.timedout)
 
-struct FDEvent
-    readable::Bool
-    writable::Bool
-    disconnect::Bool
-    timedout::Bool
-    FDEvent(r::Bool, w::Bool, d::Bool, t::Bool) = new(r, w, d, t)
-end
 # libuv file descriptor event flags
-const UV_READABLE = 1
-const UV_WRITABLE = 2
-const UV_DISCONNECT = 4
+const UV_READABLE = Int32(1)
+const UV_WRITABLE = Int32(2)
+const UV_DISCONNECT = Int32(4)
+const UV_PRIORITIZED = Int32(8)
+struct FDEvent
+    events::Int32
+    FDEvent(flags::Integer=0) = new(flags)
+end
+
+FDEvent(r::Bool, w::Bool, d::Bool, t::Bool) = FDEvent((UV_READABLE * r) | (UV_WRITABLE * w) | (UV_DISCONNECT * d)) # deprecated method
+
+function getproperty(f::FDEvent, field::Symbol)
+    events = getfield(f, :events)
+    field === :readable && return (events & UV_READABLE) != 0
+    field === :writable && return (events & UV_WRITABLE) != 0
+    field === :disconnect && return (events & UV_DISCONNECT) != 0
+    field === :prioritized && return (events & UV_PRIORITIZED) != 0
+    field === :timedout && return events == 0
+    field === :events && return Int(events)
+    getfield(f, field)::Union{}
+end
+propertynames(f::FDEvent) = (:readable, :writable, :disconnect, :prioritized, :timedout, :events)
 
 isreadable(f::FDEvent) = f.readable
 iswritable(f::FDEvent) = f.writable
-FDEvent() = FDEvent(false, false, false, true)
-FDEvent(flags::Integer) = FDEvent((flags & UV_READABLE) != 0,
-                                  (flags & UV_WRITABLE) != 0,
-                                  (flags & UV_DISCONNECT) != 0,
-                                  false)
-|(a::FDEvent, b::FDEvent) =
-    FDEvent(a.readable | b.readable,
-            a.writable | b.writable,
-            a.disconnect | b.disconnect,
-            a.timedout | b.timedout)
+|(a::FDEvent, b::FDEvent) = FDEvent(getfield(a, :events) | getfield(b, :events))
 
 mutable struct FileMonitor
-    handle::Ptr{Cvoid}
+    @atomic handle::Ptr{Cvoid}
     file::String
     notify::Base.ThreadSynchronizer
     events::Int32
@@ -93,15 +101,15 @@ mutable struct FileMonitor
     end
 end
 
-
 mutable struct FolderMonitor
-    handle::Ptr{Cvoid}
-    notify::Channel{Any} # eltype = Union{Pair{String, FileEvent}, IOError}
-    open::Bool
+    @atomic handle::Ptr{Cvoid}
+    # notify::Channel{Any} # eltype = Union{Pair{String, FileEvent}, IOError}
+    notify::Base.ThreadSynchronizer
+    channel::Vector{Any} # eltype = Pair{String, FileEvent}
     FolderMonitor(folder::AbstractString) = FolderMonitor(String(folder))
     function FolderMonitor(folder::String)
         handle = Libc.malloc(_sizeof_uv_fs_event)
-        this = new(handle, Channel(Inf), false)
+        this = new(handle, Base.ThreadSynchronizer(), [])
         associate_julia_struct(handle, this)
         iolock_begin()
         err = ccall(:uv_fs_event_init, Cint, (Ptr{Cvoid}, Ptr{Cvoid}), eventloop(), handle)
@@ -109,7 +117,6 @@ mutable struct FolderMonitor
             Libc.free(handle)
             throw(_UVError("FolderMonitor", err))
         end
-        this.open = true
         finalizer(uvfinalize, this)
         uv_error("FolderMonitor (start)",
                  ccall(:uv_fs_event_start, Int32, (Ptr{Cvoid}, Ptr{Cvoid}, Cstring, Int32),
@@ -120,7 +127,7 @@ mutable struct FolderMonitor
 end
 
 mutable struct PollingFileWatcher
-    handle::Ptr{Cvoid}
+    @atomic handle::Ptr{Cvoid}
     file::String
     interval::UInt32
     notify::Base.ThreadSynchronizer
@@ -145,16 +152,17 @@ mutable struct PollingFileWatcher
 end
 
 mutable struct _FDWatcher
-    handle::Ptr{Cvoid}
+    @atomic handle::Ptr{Cvoid}
     fdnum::Int # this is NOT the file descriptor
     refcount::Tuple{Int, Int}
     notify::Base.ThreadSynchronizer
     events::Int32
     active::Tuple{Bool, Bool}
 
-    let FDWatchers = Vector{Any}()
+    let FDWatchers = Vector{Any}() # n.b.: this structure and the refcount are protected by the iolock
         global _FDWatcher, uvfinalize
         @static if Sys.isunix()
+            _FDWatcher(fd::RawFD, mask::FDEvent) = _FDWatcher(fd, mask.readable, mask.writable)
             function _FDWatcher(fd::RawFD, readable::Bool, writable::Bool)
                 if !readable && !writable
                     throw(ArgumentError("must specify at least one of readable or writable to create a FDWatcher"))
@@ -181,7 +189,7 @@ mutable struct _FDWatcher
                     fdnum,
                     (Int(readable), Int(writable)),
                     Base.ThreadSynchronizer(),
-                    0,
+                    Int32(0),
                     (false, false))
                 associate_julia_struct(handle, this)
                 err = ccall(:uv_poll_init, Int32, (Ptr{Cvoid}, Ptr{Cvoid}, RawFD), eventloop(), handle, fd)
@@ -203,16 +211,16 @@ mutable struct _FDWatcher
                 if t.handle != C_NULL
                     disassociate_julia_struct(t)
                     ccall(:jl_close_uv, Cvoid, (Ptr{Cvoid},), t.handle)
-                    t.handle = C_NULL
+                    @atomic :monotonic t.handle = C_NULL
                 end
                 t.refcount = (0, 0)
                 t.active = (false, false)
                 @static if Sys.isunix()
-                    if FDWatchers[t.fdnum] == t
+                    if FDWatchers[t.fdnum] === t
                         FDWatchers[t.fdnum] = nothing
                     end
                 end
-                notify(t.notify, FDEvent())
+                notify(t.notify, Int32(0))
             finally
                 unlock(t.notify)
             end
@@ -222,10 +230,12 @@ mutable struct _FDWatcher
     end
 
     @static if Sys.iswindows()
+        _FDWatcher(fd::RawFD, mask::FDEvent) = _FDWatcher(fd, mask.readable, mask.writable)
         function _FDWatcher(fd::RawFD, readable::Bool, writable::Bool)
             handle = Libc._get_osfhandle(fd)
             return _FDWatcher(handle, readable, writable)
         end
+        _FDWatcher(fd::WindowsRawSocket, mask::FDEvent) = _FDWatcher(fd, mask.readable, mask.writable)
         function _FDWatcher(fd::WindowsRawSocket, readable::Bool, writable::Bool)
             if !readable && !writable
                 throw(ArgumentError("must specify at least one of readable or writable to create a FDWatcher"))
@@ -255,25 +265,38 @@ mutable struct _FDWatcher
 end
 
 mutable struct FDWatcher
-    watcher::_FDWatcher
-    readable::Bool
-    writable::Bool
     # WARNING: make sure `close` has been manually called on this watcher before closing / destroying `fd`
+    watcher::_FDWatcher
+    mask::FDEvent
     function FDWatcher(fd::RawFD, readable::Bool, writable::Bool)
-        this = new(_FDWatcher(fd, readable, writable), readable, writable)
+        return FDWatcher(fd, FDEvent(readable, writable, false, false))
+    end
+    function FDWatcher(fd::RawFD, mask::FDEvent)
+        this = new(_FDWatcher(fd, mask), mask)
         finalizer(close, this)
         return this
     end
     @static if Sys.iswindows()
         function FDWatcher(fd::WindowsRawSocket, readable::Bool, writable::Bool)
-            this = new(_FDWatcher(fd, readable, writable), readable, writable)
+            return FDWatcher(fd, FDEvent(readable, writable, false, false))
+        end
+        function FDWatcher(fd::WindowsRawSocket, mask::FDEvent)
+            this = new(_FDWatcher(fd, mask), mask)
             finalizer(close, this)
             return this
         end
     end
 end
 
+function getproperty(fdw::FDWatcher, s::Symbol)
+    # support deprecated field names
+    s === :readable && return fdw.mask.readable
+    s === :writable && return fdw.mask.writable
+    return getfield(fdw, s)
+end
 
+
+close(t::_FDWatcher, mask::FDEvent) = close(t, mask.readable, mask.writable)
 function close(t::_FDWatcher, readable::Bool, writable::Bool)
     iolock_begin()
     if t.refcount != (0, 0)
@@ -281,31 +304,39 @@ function close(t::_FDWatcher, readable::Bool, writable::Bool)
     end
     if t.refcount == (0, 0)
         uvfinalize(t)
+    else
+        @lock t.notify notify(t.notify, Int32(0))
     end
     iolock_end()
     nothing
 end
 
 function close(t::FDWatcher)
-    r, w = t.readable, t.writable
-    t.readable = t.writable = false
-    close(t.watcher, r, w)
+    mask = t.mask
+    t.mask = FDEvent()
+    close(t.watcher, mask)
 end
 
 function uvfinalize(uv::Union{FileMonitor, FolderMonitor, PollingFileWatcher})
-    disassociate_julia_struct(uv)
-    close(uv)
+    iolock_begin()
+    if uv.handle != C_NULL
+        disassociate_julia_struct(uv) # close (and free) without notify
+        ccall(:jl_close_uv, Cvoid, (Ptr{Cvoid},), uv.handle)
+    end
+    iolock_end()
 end
 
 function close(t::Union{FileMonitor, FolderMonitor, PollingFileWatcher})
+    iolock_begin()
     if t.handle != C_NULL
         ccall(:jl_close_uv, Cvoid, (Ptr{Cvoid},), t.handle)
     end
+    iolock_end()
 end
 
 function _uv_hook_close(uv::_FDWatcher)
     # fyi: jl_atexit_hook can cause this to get called too
-    uv.handle = C_NULL
+    Libc.free(@atomicswap :monotonic uv.handle = C_NULL)
     uvfinalize(uv)
     nothing
 end
@@ -313,8 +344,8 @@ end
 function _uv_hook_close(uv::PollingFileWatcher)
     lock(uv.notify)
     try
-        uv.handle = C_NULL
         uv.active = false
+        Libc.free(@atomicswap :monotonic uv.handle = C_NULL)
         notify(uv.notify, StatStruct())
     finally
         unlock(uv.notify)
@@ -325,8 +356,8 @@ end
 function _uv_hook_close(uv::FileMonitor)
     lock(uv.notify)
     try
-        uv.handle = C_NULL
         uv.active = false
+        Libc.free(@atomicswap :monotonic uv.handle = C_NULL)
         notify(uv.notify, FileEvent())
     finally
         unlock(uv.notify)
@@ -335,12 +366,22 @@ function _uv_hook_close(uv::FileMonitor)
 end
 
 function _uv_hook_close(uv::FolderMonitor)
-    uv.open = false
-    uv.handle = C_NULL
-    close(uv.notify)
+    lock(uv.notify)
+    try
+        Libc.free(@atomicswap :monotonic uv.handle = C_NULL)
+        notify_error(uv.notify, EOFError())
+    finally
+        unlock(uv.notify)
+    end
     nothing
 end
 
+isopen(fm::FileMonitor) = fm.handle != C_NULL
+isopen(fm::FolderMonitor) = fm.handle != C_NULL
+isopen(pfw::PollingFileWatcher) = pfw.handle != C_NULL
+isopen(pfw::_FDWatcher) = pfw.refcount != (0, 0)
+isopen(pfw::FDWatcher) = !pfw.mask.timedout
+
 function uv_fseventscb_file(handle::Ptr{Cvoid}, filename::Ptr, events::Int32, status::Int32)
     t = @handle_as handle FileMonitor
     lock(t.notify)
@@ -359,11 +400,17 @@ end
 
 function uv_fseventscb_folder(handle::Ptr{Cvoid}, filename::Ptr, events::Int32, status::Int32)
     t = @handle_as handle FolderMonitor
-    if status != 0
-        put!(t.notify, _UVError("FolderMonitor", status))
-    else
-        fname = (filename == C_NULL) ? "" : unsafe_string(convert(Cstring, filename))
-        put!(t.notify, fname => FileEvent(events))
+    lock(t.notify)
+    try
+        if status != 0
+            notify_error(t.notify, _UVError("FolderMonitor", status))
+        else
+            fname = (filename == C_NULL) ? "" : unsafe_string(convert(Cstring, filename))
+            push!(t.channel, fname => FileEvent(events))
+            notify(t.notify)
+        end
+    finally
+        unlock(t.notify)
     end
     nothing
 end
@@ -384,7 +431,7 @@ function uv_pollcb(handle::Ptr{Cvoid}, status::Int32, events::Int32)
                     ccall(:uv_poll_stop, Int32, (Ptr{Cvoid},), t.handle)
                 end
             end
-            notify(t.notify, FDEvent(events))
+            notify(t.notify, events)
         end
     finally
         unlock(t.notify)
@@ -416,12 +463,17 @@ function __init__()
     global uv_jl_fspollcb = @cfunction(uv_fspollcb, Cvoid, (Ptr{Cvoid}, Cint, Ptr{Cvoid}, Ptr{Cvoid}))
     global uv_jl_fseventscb_file = @cfunction(uv_fseventscb_file, Cvoid, (Ptr{Cvoid}, Ptr{Int8}, Int32, Int32))
     global uv_jl_fseventscb_folder = @cfunction(uv_fseventscb_folder, Cvoid, (Ptr{Cvoid}, Ptr{Int8}, Int32, Int32))
+
+    Base.mkpidlock_hook = mkpidlock
+    Base.trymkpidlock_hook = trymkpidlock
+    Base.parse_pidfile_hook = Pidfile.parse_pidfile
+
     nothing
 end
 
 function start_watching(t::_FDWatcher)
     iolock_begin()
-    t.handle == C_NULL && return throw(ArgumentError("FDWatcher is closed"))
+    t.handle == C_NULL && throw(ArgumentError("FDWatcher is closed"))
     readable = t.refcount[1] > 0
     writable = t.refcount[2] > 0
     if t.active[1] != readable || t.active[2] != writable
@@ -439,7 +491,7 @@ end
 
 function start_watching(t::PollingFileWatcher)
     iolock_begin()
-    t.handle == C_NULL && return throw(ArgumentError("PollingFileWatcher is closed"))
+    t.handle == C_NULL && throw(ArgumentError("PollingFileWatcher is closed"))
     if !t.active
         uv_error("PollingFileWatcher (start)",
                  ccall(:uv_fs_poll_start, Int32, (Ptr{Cvoid}, Ptr{Cvoid}, Cstring, UInt32),
@@ -468,7 +520,7 @@ end
 
 function start_watching(t::FileMonitor)
     iolock_begin()
-    t.handle == C_NULL && return throw(ArgumentError("FileMonitor is closed"))
+    t.handle == C_NULL && throw(ArgumentError("FileMonitor is closed"))
     if !t.active
         uv_error("FileMonitor (start)",
                  ccall(:uv_fs_event_start, Int32, (Ptr{Cvoid}, Ptr{Cvoid}, Cstring, Int32),
@@ -495,67 +547,78 @@ function stop_watching(t::FileMonitor)
     nothing
 end
 
-function wait(fdw::FDWatcher)
-    GC.@preserve fdw begin
-        return wait(fdw.watcher, readable = fdw.readable, writable = fdw.writable)
-    end
-end
-
-function wait(fdw::_FDWatcher; readable=true, writable=true)
-    events = FDEvent(Int32(0))
+# n.b. this _wait may return spuriously early with a timedout event
+function _wait(fdw::_FDWatcher, mask::FDEvent)
     iolock_begin()
     preserve_handle(fdw)
     lock(fdw.notify)
     try
-        while true
-            haveevent = false
-            events |= FDEvent(fdw.events)
-            if readable && isreadable(events)
-                fdw.events &= ~UV_READABLE
-                haveevent = true
-            end
-            if writable && iswritable(events)
-                fdw.events &= ~UV_WRITABLE
-                haveevent = true
-            end
-            if haveevent
-                break
-            end
-            if fdw.refcount == (0, 0) # !open
-                throw(EOFError())
-            else
-                start_watching(fdw) # make sure the poll is active
-                iolock_end()
-                events = wait(fdw.notify)::FDEvent
-                unlock(fdw.notify)
-                iolock_begin()
-                lock(fdw.notify)
-            end
+        events = FDEvent(fdw.events & mask.events)
+        if !isopen(fdw) # !open
+            throw(EOFError())
+        elseif events.timedout
+            start_watching(fdw) # make sure the poll is active
+            iolock_end()
+            return FDEvent(wait(fdw.notify)::Int32)
+        else
+            iolock_end()
+            return events
         end
     finally
         unlock(fdw.notify)
         unpreserve_handle(fdw)
     end
-    iolock_end()
-    return events
 end
 
-function wait(fd::RawFD; readable=false, writable=false)
-    fdw = _FDWatcher(fd, readable, writable)
-    try
-        return wait(fdw, readable=readable, writable=writable)
-    finally
-        close(fdw, readable, writable)
+function wait(fdw::_FDWatcher; readable=true, writable=true)
+    return wait(fdw, FDEvent(readable, writable, false, false))
+end
+function wait(fdw::_FDWatcher, mask::FDEvent)
+    while true
+        mask.timedout && return mask
+        events = _wait(fdw, mask)
+        if !events.timedout
+            @lock fdw.notify fdw.events &= ~events.events
+            return events
+        end
     end
 end
 
+function wait(fdw::FDWatcher)
+    isopen(fdw) || throw(EOFError())
+    while true
+        events = GC.@preserve fdw _wait(fdw.watcher, fdw.mask)
+        isopen(fdw) || throw(EOFError())
+        if !events.timedout
+            @lock fdw.watcher.notify fdw.watcher.events &= ~events.events
+            return events
+        end
+    end
+end
+
+function wait(socket::RawFD; readable=false, writable=false)
+    return wait(socket, FDEvent(readable, writable, false, false))
+end
+function wait(fd::RawFD, mask::FDEvent)
+    fdw = _FDWatcher(fd, mask)
+    try
+        return wait(fdw, mask)
+    finally
+        close(fdw, mask)
+    end
+end
+
+
 if Sys.iswindows()
     function wait(socket::WindowsRawSocket; readable=false, writable=false)
-        fdw = _FDWatcher(socket, readable, writable)
+        return wait(socket, FDEvent(readable, writable, false, false))
+    end
+    function wait(socket::WindowsRawSocket, mask::FDEvent)
+        fdw = _FDWatcher(socket, mask)
         try
-            return wait(fdw, readable=readable, writable=writable)
+            return wait(fdw, mask)
         finally
-            close(fdw, readable, writable)
+            close(fdw, mask)
         end
     end
 end
@@ -611,26 +674,20 @@ function wait(m::FileMonitor)
 end
 
 function wait(m::FolderMonitor)
-    m.handle == C_NULL && return throw(ArgumentError("FolderMonitor is closed"))
-    if isready(m.notify)
-        evt = take!(m.notify) # non-blocking fast-path
-    else
-        preserve_handle(m)
-        evt = try
-                take!(m.notify)
-            catch ex
-                unpreserve_handle(m)
-                if ex isa InvalidStateException && ex.state === :closed
-                    rethrow(EOFError()) # `wait(::Channel)` throws the wrong exception
-                end
-                rethrow()
+    m.handle == C_NULL && throw(EOFError())
+    preserve_handle(m)
+    lock(m.notify)
+    evt = try
+            m.handle == C_NULL && throw(EOFError())
+            while isempty(m.channel)
+                wait(m.notify)
             end
-    end
-    if evt isa Pair{String, FileEvent}
-        return evt
-    else
-        throw(evt)
-    end
+            popfirst!(m.channel)
+        finally
+            unlock(m.notify)
+            unpreserve_handle(m)
+        end
+    return evt::Pair{String, FileEvent}
 end
 
 
@@ -647,28 +704,45 @@ The returned value is an object with boolean fields `readable`, `writable`, and
 giving the result of the polling.
 """
 function poll_fd(s::Union{RawFD, Sys.iswindows() ? WindowsRawSocket : Union{}}, timeout_s::Real=-1; readable=false, writable=false)
-    wt = Condition()
-    fdw = _FDWatcher(s, readable, writable)
+    mask = FDEvent(readable, writable, false, false)
+    mask.timedout && return mask
+    fdw = _FDWatcher(s, mask)
+    local timer
+    # we need this flag to explicitly track whether we call `close` already, to update the internal refcount correctly
+    timedout = false # TODO: make this atomic
     try
         if timeout_s >= 0
-            result::FDEvent = FDEvent()
-            @async (sleep(timeout_s); notify(wt))
-            @async begin
-                try
-                    result = wait(fdw, readable=readable, writable=writable)
-                catch e
-                    notify_error(wt, e)
-                    return
+            # delay creating the timer until shortly before we start the poll wait
+            timer = Timer(timeout_s) do t
+                timedout && return
+                timedout = true
+                close(fdw, mask)
+            end
+            try
+                while true
+                    events = _wait(fdw, mask)
+                    if timedout || !events.timedout
+                        @lock fdw.notify fdw.events &= ~events.events
+                        return events
+                    end
                 end
-                notify(wt)
+            catch ex
+                ex isa EOFError() || rethrow()
+                return FDEvent()
             end
-            wait(wt)
-            return result
         else
-            return wait(fdw, readable=readable, writable=writable)
+            return wait(fdw, mask)
         end
     finally
-        close(fdw, readable, writable)
+        if @isdefined(timer)
+            if !timedout
+                timedout = true
+                close(timer)
+                close(fdw, mask)
+            end
+        else
+            close(fdw, mask)
+        end
     end
 end
 
@@ -676,37 +750,46 @@ end
     watch_file(path::AbstractString, timeout_s::Real=-1)
 
 Watch file or directory `path` for changes until a change occurs or `timeout_s` seconds have
-elapsed.
+elapsed. This function does not poll the file system and instead uses platform-specific
+functionality to receive notifications from the operating system (e.g. via inotify on Linux).
+See the NodeJS documentation linked below for details.
 
-The returned value is an object with boolean fields `changed`, `renamed`, and `timedout`,
+The returned value is an object with boolean fields `renamed`, `changed`, and `timedout`,
 giving the result of watching the file.
 
 This behavior of this function varies slightly across platforms. See
 <https://nodejs.org/api/fs.html#fs_caveats> for more detailed information.
 """
-function watch_file(s::AbstractString, timeout_s::Real=-1)
+function watch_file(s::String, timeout_s::Float64=-1.0)
     fm = FileMonitor(s)
+    local timer
     try
         if timeout_s >= 0
-            @async (sleep(timeout_s); close(fm))
+            timer = Timer(timeout_s) do t
+                close(fm)
+            end
         end
         return wait(fm)
     finally
         close(fm)
+        @isdefined(timer) && close(timer)
     end
 end
+watch_file(s::AbstractString, timeout_s::Real=-1) = watch_file(String(s), Float64(timeout_s))
 
 """
     watch_folder(path::AbstractString, timeout_s::Real=-1)
 
 Watches a file or directory `path` for changes until a change has occurred or `timeout_s`
-seconds have elapsed.
+seconds have elapsed. This function does not poll the file system and instead uses platform-specific
+functionality to receive notifications from the operating system (e.g. via inotify on Linux).
+See the NodeJS documentation linked below for details.
 
 This will continuing tracking changes for `path` in the background until
 `unwatch_folder` is called on the same `path`.
 
 The returned value is an pair where the first field is the name of the changed file (if available)
-and the second field is an object with boolean fields `changed`, `renamed`, and `timedout`,
+and the second field is an object with boolean fields `renamed`, `changed`, and `timedout`,
 giving the event.
 
 This behavior of this function varies slightly across platforms. See
@@ -717,34 +800,39 @@ function watch_folder(s::String, timeout_s::Real=-1)
     fm = get!(watched_folders, s) do
         return FolderMonitor(s)
     end
-    if timeout_s >= 0 && !isready(fm.notify)
+    local timer
+    if timeout_s >= 0
+        @lock fm.notify isempty(fm.channel) || return popfirst!(fm.channel)
         if timeout_s <= 0.010
-            # for very small timeouts, we can just sleep for the timeout-interval
+            # for very small timeouts, we can just sleep for the whole timeout-interval
             (timeout_s == 0) ? yield() : sleep(timeout_s)
-            if !isready(fm.notify)
-                return "" => FileEvent() # timeout
-            end
-            # fall-through to a guaranteed non-blocking fast-path call to wait
+            @lock fm.notify isempty(fm.channel) || return popfirst!(fm.channel)
+            return "" => FileEvent() # timeout
         else
-            # If we may need to be able to cancel via a timeout,
-            # create a second monitor object just for that purpose.
-            # We still take the events from the primary stream.
-            fm2 = FileMonitor(s)
-            try
-                @async (sleep(timeout_s); close(fm2))
-                while isopen(fm.notify) && !isready(fm.notify)
-                    fm2.handle == C_NULL && return "" => FileEvent() # timeout
-                    wait(fm2)
-                end
-            finally
-                close(fm2)
+            timer = Timer(timeout_s) do t
+                @lock fm.notify notify(fm.notify)
             end
-            # guaranteed that next call to `wait(fm)` is non-blocking
-            # since we haven't entered the libuv event loop yet
-            # or the Base scheduler workqueue since last testing `isready`
         end
     end
-    return wait(fm)
+    # inline a copy of `wait` with added support for checking timer
+    fm.handle == C_NULL && throw(EOFError())
+    preserve_handle(fm)
+    lock(fm.notify)
+    evt = try
+            fm.handle == C_NULL && throw(EOFError())
+            while isempty(fm.channel)
+                if @isdefined(timer)
+                    isopen(timer) || return "" => FileEvent() # timeout
+                end
+                wait(fm.notify)
+            end
+            popfirst!(fm.channel)
+        finally
+            unlock(fm.notify)
+            unpreserve_handle(fm)
+            @isdefined(timer) && close(timer)
+        end
+    return evt::Pair{String, FileEvent}
 end
 
 """
@@ -783,9 +871,12 @@ it is more reliable and efficient, although in some situations it may not be ava
 """
 function poll_file(s::AbstractString, interval_seconds::Real=5.007, timeout_s::Real=-1)
     pfw = PollingFileWatcher(s, Float64(interval_seconds))
+    local timer
     try
         if timeout_s >= 0
-            @async (sleep(timeout_s); close(pfw))
+            timer = Timer(timeout_s) do t
+                close(pfw)
+            end
         end
         statdiff = wait(pfw)
         if isa(statdiff[2], IOError)
@@ -795,7 +886,11 @@ function poll_file(s::AbstractString, interval_seconds::Real=5.007, timeout_s::R
         return statdiff
     finally
         close(pfw)
+        @isdefined(timer) && close(timer)
     end
 end
 
+include("pidfile.jl")
+import .Pidfile: mkpidlock, trymkpidlock
+
 end
