@@ -275,14 +275,26 @@ function download(
             (total, now) -> begin
                 bar.max = total
                 bar.current = now
-                show_progress(io, bar)
+                # Downloads.download attatches the progress indicator to the header request too
+                # which is only ~100 bytes, and will report as 0 - 100% progress immediately
+                # then dip down to 0 before the actual download starts. So we only show the
+                # progress bar once the real download starts.
+                total > 1000 && show_progress(io, bar)
             end
         end
     else
-        (total, now) -> nothing
+        nothing
+    end
+    # During precompilation don't use the default downloader as it has a
+    # 30s grace period that will cause the pending process warning on 1.10
+    # (on 1.11 the downloader is finalized before the pending process detector)
+    downloader = if ccall(:jl_generating_output, Cint, ()) != 0
+        Downloads.Downloader(grace=0)
+    else
+        nothing
     end
     try
-        Downloads.download(url, dest; headers, progress)
+        Downloads.download(url, dest; headers, progress, downloader)
     finally
         do_fancy && end_progress(io, bar)
     end
@@ -345,10 +357,21 @@ function download_verify(
     mkpath(dirname(dest))
 
     # Download the file, optionally continuing
-    f = retry(delays = fill(1.0, 3)) do
-        download(url, dest; verbose=verbose || !quiet_download)
+    attempts = 3
+    for i in 1:attempts
+        try
+            download(url, dest; verbose=verbose || !quiet_download)
+            break
+        catch err
+            @debug "download and verify failed on attempt $i/$attempts" url dest err
+            # for system errors like `no space left on device` exit after first try
+            if err isa SystemError || i == attempts
+                rethrow()
+            else
+                sleep(1)
+            end
+        end
     end
-    f()
     if hash !== nothing && !verify(dest, hash; verbose=verbose)
         # If the file already existed, it's possible the initially downloaded chunk
         # was bad.  If verification fails after downloading, auto-delete the file
