@@ -19,15 +19,18 @@ const Requires = Dict{UUID,VersionSpec}
 struct Fixed
     version::VersionNumber
     requires::Requires
+    weak::Set{UUID}
 end
-Fixed(v::VersionNumber) = Fixed(v, Requires())
+Fixed(v::VersionNumber, requires::Requires = Requires()) = Fixed(v, requires, Set{UUID}())
 
-Base.:(==)(a::Fixed, b::Fixed) = a.version == b.version && a.requires == b.requires
-Base.hash(f::Fixed, h::UInt) = hash((f.version, f.requires), h + (0x68628b809fd417ca % UInt))
+Base.:(==)(a::Fixed, b::Fixed) = a.version == b.version && a.requires == b.requires && a.weak == b.weak
+Base.hash(f::Fixed, h::UInt) = hash((f.version, f.requires, f.weak), h + (0x68628b809fd417ca % UInt))
 
 Base.show(io::IO, f::Fixed) = isempty(f.requires) ?
     print(io, "Fixed(", repr(f.version), ")") :
-    print(io, "Fixed(", repr(f.version), ",", f.requires, ")")
+    isempty(f.weak) ?
+    print(io, "Fixed(", repr(f.version), ",", f.requires, ")") :
+    print(io, "Fixed(", repr(f.version), ",", f.requires, ", weak=", f.weak, ")")
 
 
 struct ResolverError <: Exception
@@ -338,6 +341,8 @@ function greedysolver(graph::Graph)
                     return (false, Int[])
                 elseif old_v1 == spp[p1]
                     sol[p1] = v1
+                    fill!(gconstr[p1], false)
+                    gconstr[p1][v1] = true
                     push!(staged_next, p1)
                 end
             end
@@ -476,7 +481,7 @@ function enforce_optimality!(sol::Vector{Int}, graph::Graph)
         copy!(old_sol, sol)
         push!(allsols, copy(sol))
 
-        # step 1: uninstall unneded packages
+        # step 1: uninstall unneeded packages
         _uninstall_unreachable!(sol, why, graph)
 
         # setp 2: try to bump each installed package in turn
